package com.jaspersoft.jasperreports.data.elasticsearch;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.Map;

import net.sf.jasperreports.engine.JRException;

import org.apache.commons.io.IOUtils;
import org.apache.http.client.fluent.Request;
import org.apache.http.client.fluent.Response;
import org.apache.http.entity.AbstractHttpEntity;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.message.BasicHeader;
import org.apache.http.protocol.HTTP;

import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.jaspersoft.webservice.data.enumerations.AuthType;
import com.jaspersoft.webservice.data.enumerations.VerbType;
import com.jaspersoft.webservice.data.util.WebServiceConnector;

public class ElasticSearchConnector extends WebServiceConnector {

	private static final String BATCH_SIZE_PROPERTY = "batchSize";
	
	private static final String RESULT_SIZE_PROPERTY = "resultSize";
	
	private static final String USE_SCAN_PROPERTY = "useScan";
	
	private static final String SCAN_TIME_PROPERTY = "scanTime";
	
	private static final String QUERY_PROPERTY = "query";
	
	private static ObjectMapper mapper = null;
	
	Long resultSize;
	// Default batch size
	Integer batchSize = 50;
	Boolean useScan = true;
	
	String nextScrollId;
	String scanTime = "1m"; //default to 1 minute
	Long totalHits = 0L;
	Long totalResults = 0L;
	Integer batchNumber = 0;
	String searchQuery;
	Integer shards;
	
	private Records topResult;
	boolean firstNext = true;
	private String queryUrl;
	
	public static ObjectMapper getObjectMapper() {
		if (mapper == null) {
			mapper = new ObjectMapper();
			
			mapper.configure(JsonParser.Feature.ALLOW_UNQUOTED_FIELD_NAMES, true);
			mapper.configure(JsonParser.Feature.ALLOW_SINGLE_QUOTES, true);
			mapper.configure(JsonParser.Feature.ALLOW_COMMENTS, true);
		}
		return mapper;
	}
	
	protected ElasticSearchConnector(){
		super();
	}

	@Override
	public InputStream callWS(String wsURI, String baseURL, VerbType verbType, 
			  AuthType authType, String contentType, String content,
			  Map<String, String> authMap, Map<String, String> headerMap, 
			  Map<String,String> GETParameters) throws JRException {
		
		//Initialize the fields
		this.verbType = verbType;
		this.content = content;
		this.contentType = contentType;
		this.authType = authType;
		this.wsURI = wsURI;
		this.baseURL = baseURL;
		this.authMap = authMap;
		this.headerMap = headerMap;
		this.GETParameters = GETParameters;
		
		//do the request
		this.topResult = executeQuery();
		this.totalResults = topResult.getResultSize();

		while(topResult != null && getResultSize() > getTotalResults()){
			Records nextBatch = executeNextResults();
			if (nextBatch != null && nextBatch.getResultSize() > 0) {
				topResult = topResult.merge(nextBatch.getResultNode());
				setTotalResults(getTotalResults() + nextBatch.getResultSize());
			}
		}

		if (topResult != null){
			try {
				Object obj = getObjectMapper().treeToValue(topResult.getResultNode(), Object.class);
				String json = getObjectMapper().writeValueAsString(obj);
				InputStream stream = IOUtils.toInputStream(json);
				return stream;
			} catch (JsonProcessingException e) {
				throw new JRException(e);
			}
		} else {
			return null;
		}
	}
	
	public Records executeQuery() throws JRException {
		
		//Set the request additional parameters
		if (headerMap.containsKey(RESULT_SIZE_PROPERTY)) {
			setResultSize(Long.parseLong(headerMap.get(RESULT_SIZE_PROPERTY)));
		}

		if (headerMap.containsKey(BATCH_SIZE_PROPERTY)) {
			setBatchSize(Integer.parseInt(headerMap.get(BATCH_SIZE_PROPERTY)));
		}
		
		if (headerMap.containsKey(USE_SCAN_PROPERTY)) {
			setUseScan(Boolean.parseBoolean(headerMap.get(USE_SCAN_PROPERTY)));
		}
		
		if (isUseScan() && headerMap.containsKey(SCAN_TIME_PROPERTY)) {
			setScanTime(headerMap.get(SCAN_TIME_PROPERTY));
		}
		
		//Get the url
		String queryUrl = getRequestUrl();
		setQueryUrl(queryUrl);
		
		//Create the request
		JsonNode queryRequest;
		try {
			String requestContent = content.trim().isEmpty() ? "{}" : content.trim(); 
			queryRequest = getObjectMapper().readTree(requestContent);
		} catch (IOException e) {
			throw new JRException(e);
		}
		Records parsedQueryRequest = new Records(queryRequest);

		JsonNode query = (JsonNode) parsedQueryRequest.getObject(QUERY_PROPERTY);
		if (query != null){
			setSearchQuery(query.toString());
		} else {
			setSearchQuery(queryRequest.toString());
		}

		//Generate the request url
		StringBuilder parameters = new StringBuilder("?");
		try{
			if (isUseScan()) {
				parameters.append(URLEncoder.encode("search_type", ENCODING));
				parameters.append("="); //$NON-NLS-1$
				parameters.append(URLEncoder.encode("scan", ENCODING));
				parameters.append("&"); //$NON-NLS-1$
				parameters.append(URLEncoder.encode("scroll", ENCODING));
				parameters.append("="); //$NON-NLS-1$
				parameters.append(URLEncoder.encode(getScanTime(), ENCODING));
				parameters.append("&"); //$NON-NLS-1$
				parameters.append(URLEncoder.encode("size", ENCODING));
				parameters.append("="); //$NON-NLS-1$
				parameters.append(URLEncoder.encode(getScrollBatchSize().toString(), ENCODING));
			} else {		
				parameters.append(URLEncoder.encode("from", ENCODING));
				parameters.append("=");
				parameters.append(URLEncoder.encode("0", ENCODING));
				parameters.append("&");
				parameters.append(URLEncoder.encode("size", ENCODING));
				parameters.append("=");
				parameters.append(URLEncoder.encode(getBatchSize().toString(), ENCODING));
			}
		} catch (UnsupportedEncodingException e) {
			throw new JRException(e);
		}
		
		String postURL = queryUrl+parameters.toString();
		Request req = Request.Post(postURL);
		
		if (isUseScan()) {
			req.addHeader("search_type", "scan");
			req.addHeader("scroll", getScanTime());
			req.addHeader("size", getScrollBatchSize().toString());
		} else {
			req.addHeader("from", "0");
			req.addHeader("size", getBatchSize().toString());
		}
		
		setHeaders(req);		
		setAuthenticationHeader(req);

		// Post the query as the body
		AbstractHttpEntity entity;
		try {
			entity = new ByteArrayEntity(getSearchQuery().getBytes(ENCODING));
		} catch (UnsupportedEncodingException e) {
			throw new JRException("Unsupported encoding for request! ",e);
		}
	    entity.setContentType(new BasicHeader(HTTP.CONTENT_TYPE, contentType));
	    req.body(entity);
		
		JsonNode response = null;
		try {
			Response resp = req.execute();
			response = getObjectMapper().readTree(resp.returnContent().asStream());
		} catch (Exception e1) {
			throw new JRException(e1);
		}

		firstNext = true;

		if (isUseScan()) {
			// should have gotten a scroll_result
			Records parsedScrollResponse = new Records(response);
			String scrollId = parseScrollResult(parsedScrollResponse);
			if (scrollId == null) {
				setBatchNumber(1);
				return parsedScrollResponse;
			}
			else {
				// get a first result
				return nextScrollResult(scrollId);
			}
		} else {
			setBatchNumber(1);
			return new Records(response);
		}
	}
	
	@Override
	protected String getRequestUrl() {
		String requestURL = super.getRequestUrl().trim();
		if (!requestURL.endsWith("_search")){
			if (!requestURL.endsWith("/")) requestURL += "/";
			requestURL += "_search";
		}
		return requestURL;
	}
	


	public Records executeNextResults() throws JRException {
		if (isUseScan()) {
			return nextScrollResult(getNextScrollId());
		}
		// else we are using a size window and not using scroll

		String fromValue = new Integer(getBatchSize() * getBatchNumber()).toString();
		
		StringBuilder parameters = new StringBuilder("?");
		try{
			parameters.append(URLEncoder.encode("from", ENCODING));
			parameters.append("=");
			parameters.append(URLEncoder.encode(fromValue, ENCODING));
			parameters.append("&");
			parameters.append(URLEncoder.encode("size", ENCODING));
			parameters.append("=");
			parameters.append(URLEncoder.encode(getBatchSize().toString(), ENCODING));
		} catch (UnsupportedEncodingException e) {
			throw new JRException(e);
		}
		String postURL = getQueryUrl() + parameters.toString();
		
		Request req = Request.Post(postURL);
		req.addHeader("from", fromValue);
		req.addHeader("size", getBatchSize().toString());
	
		// Post the query as the body
		AbstractHttpEntity entity;
		try {
			entity = new ByteArrayEntity(getSearchQuery().getBytes(ENCODING));
		} catch (UnsupportedEncodingException e) {
			throw new JRException("Unsupported encoding for request! ",e);
		}
	    entity.setContentType(new BasicHeader(HTTP.CONTENT_TYPE, contentType));
	    req.body(entity);
	    
	    setHeaders(req);		
	    setAuthenticationHeader(req);
	    
		JsonNode response = null;
		try {
			Response resp = req.execute();
			response = getObjectMapper().readTree(resp.returnContent().asStream());
		} catch (Exception e1) {
			throw new JRException(e1);
		} 
		setBatchNumber(getBatchNumber() + 1);
		return new Records(response);
	}
	
	private Records nextScrollResult(String scrollId) throws JRException {
		String url = wsURI.trim() + "/_search/scroll";
		
		//Load parameters and encode them
		StringBuilder urlParameters = new StringBuilder();
		try {
			urlParameters.append(URLEncoder.encode("scroll_id", ENCODING));
			urlParameters.append("="); //$NON-NLS-1$
			urlParameters.append(URLEncoder.encode(scrollId, ENCODING));
			urlParameters.append("&"); //$NON-NLS-1$
			urlParameters.append(URLEncoder.encode("scroll", ENCODING));
			urlParameters.append("="); //$NON-NLS-1$
			urlParameters.append(URLEncoder.encode(getScanTime(), ENCODING));
			urlParameters.append("&"); //$NON-NLS-1$
			urlParameters.append(URLEncoder.encode("size", ENCODING));
			urlParameters.append("="); //$NON-NLS-1$
			urlParameters.append(URLEncoder.encode(getScrollBatchSize().toString(), ENCODING));
		} catch (UnsupportedEncodingException e) {
			throw new JRException(e);
		}
		
		String fromClause = urlParameters.toString();
		if(!url.contains("?")){
			url += "?";
		} else if (!url.endsWith("?") && !url.endsWith("&")){
			url += "&";
		}
		url += fromClause;
		Request req = Request.Get(url);
		JsonNode response = null;
		try {
			Response resp = req.execute();
			response = getObjectMapper().readTree(resp.returnContent().asStream());
		} catch (Exception e1) {
			throw new JRException(e1);
		} 
		
		Records r =  new Records(response);
		// , this
		if (r != null) {
			String newScrollId = (String) r.getObject("_scroll_id", String.class);
			setNextScrollId(newScrollId);
		}
		return r;
	}

	private String parseScrollResult(Records parsedScrollResponse) throws JRException {
		//_scroll_id
		String scrollId = (String) parsedScrollResponse.getObject("_scroll_id", String.class);
		if (scrollId == null) {
			System.out.println("no _scroll_id, fallback");
			return null;
		}
		//"hits"."total":1000
		Long totalHits = (Long) parsedScrollResponse.getObject("hits.total", Long.class);
		if (totalHits == null) {
			throw new JRException("no hits.total");
		} else {
			setTotalHits(totalHits);
		}
		
		Integer shards = (Integer) parsedScrollResponse.getObject("_shards.total", Integer.class);
		if (shards == null) {
			throw new JRException("no _shards.total");
		} else {
			setShards(shards);
		}
		return scrollId;
	}

	public Integer getBatchSize() {
		return batchSize;
	}

	public Integer getScrollBatchSize() {
		if (getShards() == null) {
			return getBatchSize();
		} else {
			return (getBatchSize() / getShards()) + 1;
		}
	}
	
	public void setBatchSize(Integer batchSize) {
		this.batchSize = batchSize;
	}

	public Boolean isUseScan() {
		return useScan;
	}

	public void setUseScan(Boolean useScan) {
		this.useScan = useScan;
	}

	public String getScanTime() {
		return scanTime;
	}

	public void setScanTime(String scanTime) {
		this.scanTime = scanTime;
	}

	public Records getTopResult() {
		return topResult;
	}

	public void setTopResult(Records topResult) {
		this.topResult = topResult;
	}

	public boolean isFirstNext() {
		return firstNext;
	}

	public void setFirstNext(boolean firstNext) {
		this.firstNext = firstNext;
	}

	public Long getTotalHits() {
		return totalHits;
	}

	public void setTotalHits(Long totalHits) {
		this.totalHits = totalHits;
	}


	public String getSearchQuery() {
		return searchQuery;
	}

	public void setSearchQuery(String searchQuery) {
		this.searchQuery = searchQuery;
	}

	public Integer getBatchNumber() {
		return batchNumber;
	}

	public void setBatchNumber(Integer batchNumber) {
		this.batchNumber = batchNumber;
	}

	public String getQueryUrl() {
		return queryUrl;
	}

	public void setQueryUrl(String queryUrl) {
		this.queryUrl = queryUrl;
	}

	/**
	 * If we have not set a result size, then use the total hits
	 * @return
	 */
	public Long getResultSize() {
		if (resultSize == null) {
			return getTotalHits();
		}
		return resultSize;
	}

	public void setResultSize(Long resultSize) {
		this.resultSize = resultSize;
	}

	public String getNextScrollId() {
		return nextScrollId;
	}

	public void setNextScrollId(String nextScrollId) {
		this.nextScrollId = nextScrollId;
	}

	public Integer getShards() {
		return shards;
	}

	public void setShards(Integer shards) {
		this.shards = shards;
	}

	public Long getTotalResults() {
		return totalResults;
	}

	public void setTotalResults(Long totalResults) {
		this.totalResults = totalResults;
	}
}
