package com.jaspersoft.jasperreports.data.elasticsearch;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import net.sf.jasperreports.engine.JRException;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;

public class Records {
	
	private JsonNode resultNode;
	
	private List<Record> recordsCache = new ArrayList<Record>();
	
	private Long resultSize = null;
	
	Record currentRecord = null;
	
	private int currentPositionInResults= 0;
	
	// When constructed
	private Boolean initialState = true;

	public Records(JsonNode resultNode) throws JRException {
		this.currentPositionInResults = 0;
		this.resultNode = resultNode;
		if (resultNode.isArray()) {
			addToResults((ArrayNode) resultNode);
		} else {
			ArrayNode arrayNode = ElasticSearchConnector.getObjectMapper().createArrayNode();
			arrayNode.add(resultNode);
			addToResults(arrayNode);
		}
	}

	public Record getRecord(Integer pos) throws JRException {
		if (pos >= getRecordsCache().size()) {
			throw new JRException("invalid index in result: " + pos + ", max size: " + getRecordsCache().size());
		} else {
			return getRecordsCache().get(pos);
		}
	}
	public void addToResults(List<Record> newResults) throws JRException {
		getRecordsCache().addAll(newResults);
	}
	
	public void addToResults(ArrayNode newResults) throws JRException {
		for (JsonNode jo : newResults) {
			Record r = new Record(jo, this);
			getRecordsCache().add(r);
		}
	}

	public Long getResultSize() {
		if (resultSize == null) {
			resultSize = 0L;
			for (Record r: getRecordsCache()) {
				resultSize += r.getResultSize();
			}
		}
		return resultSize;
	}

	public int getCurrentPositionInResults() {
		return currentPositionInResults;
	}

	public void setCurrentPositionInResults(int currentPositionInResults) {
		this.currentPositionInResults = currentPositionInResults;
		currentRecord = null;
	}

	public Boolean getInitialState() {
		return initialState;
	}

	public void setInitialState(Boolean initialState) {
		this.initialState = initialState;
	}

	public List<Record> getRecordsCache() {
		return recordsCache;
	}

	public void setRecordsCache(List<Record> recordsCache) {
		this.recordsCache = recordsCache;
	}

	public boolean isFinished() {
		return (getCurrentPositionInResults() >= getRecordsCache().size());
	}
	
	public void moveFirst() {
		setCurrentPositionInResults(0);
	}
	
	public Record getCurrentRecord() throws JRException {
		if (isFinished()) {
			return null;
		} else if (currentRecord == null) {
			currentRecord = getRecord(getCurrentPositionInResults());
		}
		return currentRecord;
	}
	
	public Object getObject(String expression) throws JRException {
		if (getCurrentRecord() != null)
			return getCurrentRecord().getObject(expression);
		else
			return null;
	}
	
	public Object getObject(String expression, Class<?> valueClass) throws JRException {
		JsonNode node = (JsonNode) getObject(expression);
		if (node == null) {
			return null;
		}
		Object value;
		try {
			if (valueClass.equals(String.class)) {
				value = node.asText();

			} else if (valueClass.equals(Boolean.class)) {
				value = node.asBoolean();

			} else if (valueClass.equals(Double.class)) {
				value = node.asDouble();

			} else if (valueClass.equals(Integer.class)) {
				value = node.asInt();

			} else if (valueClass.equals(Long.class)) {
				value = node.asLong();

			} else {
				throw new JRException("Field '" + expression + "' is of class '" + valueClass.getName() + "' and cannot be converted");
			}
		} catch (Exception e) {
			throw new JRException("Unable to get value for field '" + expression + "' of class '" + valueClass.getName() + "'", e);
		}
		return value;
	}
	
	public void nextPosition() throws JRException {
		setCurrentPositionInResults(getCurrentPositionInResults() + 1);
	}
	
	/**
	 * I only increment when all my underlying Results are exhausted
	 * @return
	 * @throws JRException 
	 */
	public boolean next() throws JRException {
		
		if (isFinished()) {
			return false;
		}
		
		// ask underlying records to increment
		if (getCurrentRecord().next()) {
			setInitialState(false);
			return true;
		}

		// if underlying records are exhausted, I will increment
		
		if (!isFinished()) {
			nextPosition();
		}
		
		return !isFinished();
	}
	
	public Records merge(JsonNode updateNode) throws JRException{
		return new Records(merge(resultNode, updateNode));
	}
	
	protected JsonNode merge(JsonNode mainNode, JsonNode updateNode) {

	    Iterator<String> fieldNames = updateNode.fieldNames();
	    while (fieldNames.hasNext()) {

	        String fieldName = fieldNames.next();
	        JsonNode jsonNode = mainNode.get(fieldName);
	        // if field exists and is an embedded object
	        if (jsonNode != null && jsonNode.isObject()) {
	            merge(jsonNode, updateNode.get(fieldName));
	        }
	        else {
	            if (mainNode instanceof ObjectNode) {
	                // Overwrite field
	                JsonNode value = updateNode.get(fieldName);
	                ((ObjectNode) mainNode).set(fieldName, value);
	            }
	        }

	    }

	    return mainNode;
	}
	
	public JsonNode getResultNode(){
		return resultNode;
	}

	private class Record {
		private JsonNode jsonObj;

		/*
		 * Using a LinkedHashMap means that the subResults match the order in
		 * the JSON object, so we iterate through the result in that order down
		 * the hierarchy
		 * 
		 * Algorithm:
		 * records
				record[a]
					records[1]
						record[a]
						record[b]
					records[2]
						record[a]
				record[b]
					records[3]
						record[a]
						record[b]
					records[4]
						record[a]
						record[b]
					
			
			first
				a 1a 2a
			next
				exhausted 2 -> reset 2, increment 1
				a 1b 2a
			next
				exhausted 2, exhausted 1 - increment a
				b 3a 4a 
			next
				increment 4
				b 3a 4b
			next
				exhausted 4 -> reset 4, increment 3
				b 3b 4a
			next
				increment 4
				b 3b 4b
			next - done!
		 */
		private Map<String, Records> subResults = new LinkedHashMap<String, Records>();
		private List<Records> subResultsList;
		private Long thisResultSize = 0L;
		private Long subResultSize = null;

		Record(JsonNode jo, Records parent) throws JRException {
			this.jsonObj = jo;

			Iterator<String> it = jo.fieldNames();
			while (it.hasNext()) {
				String key = it.next();
				Object o = jo.get(key);
				//log.debug(key + ": "  + " - " + o.getClass()  + " - "  + o);
				if (o instanceof ObjectNode || o instanceof ArrayNode) {
					//log.debug(key + " is subResult");
					if (key.equalsIgnoreCase("hits") && o instanceof ArrayNode) {
						// should be an ArrayNode
						ArrayNode ar = (ArrayNode) o;
						thisResultSize = (long) ar.size();
						//log.debug("----> hits size: " + ar.size());
					}
					Records r = new Records((JsonNode) o);
					
					subResults.put(key, r);
				}
			}
			if (subResults.size() > 0) {
				subResultsList = new ArrayList<Records>(subResults.size());
				subResultsList.addAll(subResults.values());
			}
		}

		public Long getResultSize() {
			if (subResultSize == null) {
				Long subResultsResultSize = 0L;
				if (subResults.size() > 0) {
					Iterator<Records> it = subResultsList.iterator();
					while (it.hasNext()) {
						Records r = it.next();
						subResultsResultSize += (r.getResultSize() != null ? r.getResultSize() : 0L);
					}
				}
				subResultSize = subResultsResultSize;
			}
			//log.debug("getResultSize: " + (thisResultSize + subResultSize));
			return thisResultSize + subResultSize;
		}

		public Object getObject(String expression) throws JRException {
			return getObject(jsonObj, expression);
		}

		/**
		 * name.name.name down the internal results
		 * 
		 * at the current position in the results
		 * 
		 * @param expression
		 * @return
		 * @throws JRException
		 */
		public Object getObject(JsonNode currNode, String expression)
				throws JRException {
			if (expression == null || expression.length() == 0) {
				throw new JRException("empty expression");
			}
			// key at this level is string before "."
			int periodPos = expression.indexOf('.');
			String key = null;
			if (periodPos == -1) {
				key = expression;
			} else {
				key = expression.substring(0, periodPos);
			}

			Object o = currNode.get(key);
			if (o == null) {
				return null;
				//throw new JRException("No object: " + key);
			}

			if (periodPos != -1) {
				// go to the next level
				String nextLevel = expression.substring(periodPos + 1);

				if (o instanceof ArrayNode || o instanceof ObjectNode) {
					// it will be a Result

					// have we already got it?

					Records r = subResults.get(key);
					if (r == null) {
						throw new JRException("For expression: " + expression + " - no results for next level: " + key);
					}
					return r.getObject(nextLevel);
				} else {
					return null;
					/*
					 * throw new JRException("no results below: " + key +
					 * " and asking for " + nextLevel);
					 */}
			} else {
				return o;
			}
		}

		/**
		 * I only increment when all my underlying Results are exhausted
		 * 
		 * I don't have underlying results myself - I am a Record!
		 * 
		 * @return whether we can go to the next row
		 * @throws JRException
		 */
		public boolean next() throws JRException {
			if (subResultsList == null || subResultsList.isEmpty()) {
				return false;
			}
			
			// ask underlying records to increment in reverse order of declaration in the query
			ListIterator<Records> li = subResultsList.listIterator(subResultsList.size());

			while (li.hasPrevious()) {
				Records r = li.previous();
				if (!r.isFinished() && r.next()) {
					return true;
				} else {
					r.moveFirst();
				}
			}

			// otherwise I am exhausted
			return false;
		}
	}
	
}
