/*
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * Licensed under commercial Jaspersoft Subscription License Agreement
 */

package com.jaspersoft.jrs.data.webservice;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.jasperreports.data.DataAdapter;
import net.sf.jasperreports.data.DataAdapterService;
import net.sf.jasperreports.engine.DefaultJasperReportsContext;
import net.sf.jasperreports.engine.JRParameter;
import net.sf.jasperreports.engine.JasperReportsContext;
import net.sf.jasperreports.engine.design.JRDesignDataset;
import net.sf.jasperreports.engine.design.JRDesignField;
import net.sf.jasperreports.engine.design.JRDesignQuery;

import com.jaspersoft.jasperserver.api.engine.jasperreports.util.CustomDomainMetaDataImpl;
import com.jaspersoft.jasperserver.api.engine.jasperreports.util.DataAdapterDefinition;
import com.jaspersoft.jasperserver.api.engine.jasperreports.util.DataAdapterDefinitionUtil;
import com.jaspersoft.jasperserver.api.metadata.jasperreports.domain.CustomDomainMetaData;
import com.jaspersoft.jasperserver.api.metadata.jasperreports.domain.CustomReportDataSource;
import com.jaspersoft.webservice.data.WebServiceDataAdapter;
import com.jaspersoft.webservice.data.WebServiceDataAdapterImpl;
import com.jaspersoft.webservice.data.WebServiceDataAdapterService;
import com.jaspersoft.webservice.data.enumerations.AuthType;
import com.jaspersoft.webservice.data.enumerations.LanguageType;


/**
 * @author gnatali
 * custom report data source definition for WebService Query
 * This class is meant to be instantiated as a spring bean that registers a custom data source with the system.
 */
@SuppressWarnings({ "rawtypes", "unchecked" })
public class WebServiceDataSourceDefinition extends DataAdapterDefinition {
	//Currently this property value is wrong. using a constant as a workaround for the moment.
	//uncomment when the bug is fixed 
	//private static final String QUERY_LANGUAGE = WebServiceQueryExecuter.CANONICAL_LANGUAGE;
	private static final String QUERY_LANGUAGE = "WebServiceQuery";
    /**
	 * 
	 */
	private static final long serialVersionUID = 6288165536549650431L;

	public WebServiceDataSourceDefinition() {
        // add additional field
        Set<String> additionalPropertySet = getAdditionalPropertySet();
        Collection<AuthType> types = AuthType.values(DefaultJasperReportsContext.getInstance());
        for(AuthType type : types){
        	for(String parameter : type.getAuthenticator().getRequiredParameters()){
        		additionalPropertySet.add(parameter);
        	}
        }
        
        // define default values for the following properties
        Map<String, String> propertyDefaultValueMap = getPropertyDefaultValueMap();
        propertyDefaultValueMap.put("authType", "none");
        propertyDefaultValueMap.put("language", "json");
        propertyDefaultValueMap.put("verb", "post");

        // hide the following properties from UI
        Set<String> hiddenPropertySet = getHiddenPropertySet();
        hiddenPropertySet.add("authParameters");
        hiddenPropertySet.add("name");
        
        // set query executor factory
        Map<String, String> queryExecuterMap = new HashMap<String, String>();
        queryExecuterMap.put(WebServiceDataSourceDefinition.QUERY_LANGUAGE, "com.jaspersoft.webservice.data.query.WebServiceQueryExecuterFactory");
        setQueryExecuterMap(queryExecuterMap);
    }

    @Override
    public Map<String, Object>  customizePropertyValueMap(CustomReportDataSource customReportDataSource, Map<String, Object>  propertyValueMap) {
    	String authType = "none";
    	if (propertyValueMap.get("authType") != null) {
        	authType = ((String)propertyValueMap.get("authType")).toLowerCase();
        }
    	AuthType auth = AuthType.valueOf(authType, DefaultJasperReportsContext.getInstance());
    	//Use none as fallback
    	if (auth == null) auth = AuthType.NONE;
    	
    	propertyValueMap.put("authType", auth.toString());
    	
    	if (propertyValueMap.get("language") != null) {
    		String language = (String)propertyValueMap.get("language");
    		propertyValueMap.put("language", LanguageType.valueOf(language));
        }
    	
    	if (propertyValueMap.get("verb") != null) {
    		String verb = (String)propertyValueMap.get("verb");
    		propertyValueMap.put("verb", verb);
        }    	
    
    	//Transfer the parameters
    	Map<String,Object> authProps = new HashMap<String, Object>();
    	for(String parameter : auth.getAuthenticator().getRequiredParameters()){
    		String propVal = (String)propertyValueMap.get(parameter);
    		authProps.put(parameter, propVal);
    		propertyValueMap.remove(parameter);
    	}
    	propertyValueMap.put("authParameters", authProps);
        
        return propertyValueMap;
    }

    /*
    * This function is used for retrieving the metadata layer of the custom data source in form of CustomDomainMetaData
    * CustomDomainMetaData contains information JRFields, query, query language and field name mapping (actual JRField name, name used in domain)
    */
    @Override
    public CustomDomainMetaData getCustomDomainMetaData(CustomReportDataSource customDataSource) throws Exception {

        // METADATA DISCOVERY
        Map<String, Object> propertyValueMap = new HashMap<String, Object>();
        propertyValueMap = getDataSourceServicePropertyMap(customDataSource, propertyValueMap);
        propertyValueMap = customizePropertyValueMap(customDataSource, propertyValueMap);

        // create MongoDB DATA ADAPTER from Custom Report Data Source properties
        // map custom report data source properties to data adapter
        
        WebServiceDataAdapter webServiceDataAdapter = new WebServiceDataAdapterImpl();
        webServiceDataAdapter = (WebServiceDataAdapter) setupDataAdapter(webServiceDataAdapter, propertyValueMap);

        if (getValidator() != null) getValidator().validatePropertyValues(customDataSource, null);
        String query = (String)propertyValueMap.get("query");
        List<JRDesignField> jrDesignFields = getJRDesignFields(webServiceDataAdapter, query);

        List<String> columnNames = new ArrayList<String>();
        List<String> columnTypes = new ArrayList<String>();
        List<String> columnDescriptions = new ArrayList<String>();
        for (JRDesignField field : jrDesignFields) {
            String designName = field.getName().replace(".", "_");
            columnNames.add(designName);
            if (isSupportedType(field.getValueClassName())) columnTypes.add(field.getValueClassName());
            else columnTypes.add("java.lang.String");    // if it is not supported types, converts it to string for now
            columnDescriptions.add(field.getName());
        }
        // create CustomDomainMetaDataImpl object
        CustomDomainMetaDataImpl sourceMetadata = new CustomDomainMetaDataImpl();
        sourceMetadata.setQueryLanguage(WebServiceDataSourceDefinition.QUERY_LANGUAGE);
        sourceMetadata.setFieldNames(columnNames);
        Map<String, String> fieldMapping = new HashMap<String, String>();
        for (String str : columnNames) fieldMapping.put(str, str);
        sourceMetadata.setFieldMapping(fieldMapping);
        // set default column data type based on the actual data
        sourceMetadata.setFieldTypes(columnTypes);
        sourceMetadata.setQueryText(query);
        sourceMetadata.setFieldDescriptions(columnDescriptions);

        return sourceMetadata;

    }

    /**
	*  Return data adapter service for this custom data source
	**/	
    @Override
    public DataAdapterService getDataAdapterService(JasperReportsContext jasperReportsContext, DataAdapter dataAdapter) {
            return new WebServiceDataAdapterService(jasperReportsContext, (WebServiceDataAdapter)dataAdapter);
    }

    //This is just a draft, we'll have to work on this
    protected List<JRDesignField> getJRDesignFields(WebServiceDataAdapter dataAdapter, String query) throws Exception {
            Map<String,Object> parameterValues = new HashMap<String, Object>();
            // convert your param map to fill params...
            Map<String, ? extends JRParameter>  fillParams = DataAdapterDefinitionUtil.convertToFillParameters(parameterValues, getQueryExecuterFactory().getBuiltinParameters());
            //MongoDbConnection mongoDbConnection = new MongoDbConnection(dataAdapter.getMongoURI(), dataAdapter.getUsername(), dataAdapter.getPassword());
            JRDesignDataset designDataset = new JRDesignDataset(false);
            JRDesignQuery jrquery = new JRDesignQuery();
            jrquery.setText(query);
            jrquery.setLanguage(WebServiceDataSourceDefinition.QUERY_LANGUAGE);
            designDataset.setQuery(jrquery);
            //return WebServiceFieldsProvider.getInstance().getFields(DataAdapterDefinitionUtil.getJasperReportsContext(), designDataset, fillParams);
            return new ArrayList<JRDesignField>();
    }

	private static Set supportedTypeSet;
    static {
        supportedTypeSet = new HashSet();
        supportedTypeSet.add("java.lang.String");
        supportedTypeSet.add("java.lang.Byte");
        supportedTypeSet.add("java.lang.Short");
        supportedTypeSet.add("java.lang.Integer");
        supportedTypeSet.add("java.lang.Long");
        supportedTypeSet.add("java.lang.Float");
        supportedTypeSet.add("java.lang.Double");
        supportedTypeSet.add("java.lang.Number");
        supportedTypeSet.add("java.util.Date");
        supportedTypeSet.add("java.sql.Date");
        supportedTypeSet.add("java.sql.Time");
        supportedTypeSet.add("java.sql.Timestamp");
        supportedTypeSet.add("java.math.BigDecimal");
        supportedTypeSet.add("java.math.BigInteger");
        supportedTypeSet.add("java.lang.Boolean");
        supportedTypeSet.add("java.lang.Object");
    }

    private boolean isSupportedType(String type) {
        return supportedTypeSet.contains(type);
    }

}

