/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.enumerations;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.jaspersoft.webservice.data.authenitcator.AuthenticatorInterface;
import com.jaspersoft.webservice.data.authenitcator.BasicAuthenticator;
import com.jaspersoft.webservice.data.authenitcator.EmptyAuthenticator;

import net.sf.jasperreports.engine.JasperReportsContext;

/**
 * Class used to provide the authentication types. It work in a similar way to an enum
 * but it is a class to be easily serialized and extensible in the future trough the JR
 * extensions. The object of this class can't be build directly, the class must be extended.
 * It also bind each type to an authenticator that can handle that type
 * 
 * @author Orlandin Marco
 *
 */
public class AuthType {
	
	/**
	 * The logger
	 */
	protected static final Log logger = LogFactory.getLog(AuthType.class);
	
	/**
	 * The map where all the types instantiated (also if it is a subclass) are stored 
	 */
    protected static final Map<String, AuthType> KINDS = new HashMap<String, AuthType>();
    
    /**
     * Map with the authenticator for each authentication type
     */
    protected static final HashMap<AuthType, AuthenticatorInterface> authenticators = new HashMap<AuthType, AuthenticatorInterface>();

    /**
     * The static authentication type NONE, it dosen't do any authentication
     */
    public static final AuthType NONE = new AuthType("none", new EmptyAuthenticator());
    
    /**
     * The static authentication type basic, it does an authentication based on username and password
     */
    public static final AuthType BASIC = new AuthType("basic", new BasicAuthenticator());

    /**
     * The value of the current type
     */
    protected final String value;
    
    /**
     * Create a new type
     * 
     * @param value  The value of the current type
     * @param authenticator the authenticator for the type, shouldn't be null
     */
    protected AuthType(final String value, AuthenticatorInterface authenticator) {
    	if (KINDS.containsKey(value)) {
    		throw new RuntimeException("Authenticator for "+value+" already defined");
    	}
    	this.value = value;
        KINDS.put(value, this);
    	authenticators.put(this, authenticator);
    }

    /**
     * Get an authentication type from its textual value
     * 
     * @param kind the textual representation of the type
     * @return the AuthType associated or null if the name doesn't correspond to
     * any authentication 
     */
    public static AuthType valueOf(final String kind) {
	    return KINDS.get(kind);
    }

    /**
     * Get an authentication type from its textual value. If it is not found 
     * try to search it between the context extensions
     * 
     * @param kind the textual representation of the type
     * @param context a jasperreports context to search the authenticator between the extensions
     * if at first it can't be found
     * @return the AuthType associated or null if the name doesn't correspond to
     * any authentication 
     */
    public static AuthType valueOf(final String kind, JasperReportsContext context) {
	    AuthType result = valueOf(kind);
	    //Not found, try to search it inside the extensions
	    if (result == null){
	    	context.getExtensions(AuthType.class);
	    	result = valueOf(kind);
	    }
	    return result;
    }
    
    /**
     * Return all the defined authentication types
     * 
     * @return a not null list of the textual representation of the authentication types
     */
    public static Collection<AuthType> values(){
    	return KINDS.values();
    }
    
    /**
     * Return all the defined authentication types. Before try to load all the authentication
     * provided by the extension
     * 
     * @return a not null list of the textual representation of the authentication types
     *
     */
    public static Collection<AuthType> values(JasperReportsContext context) {
    	context.getExtensions(AuthType.class);
    	return values();
    }
    
    /**
     * Return the current type as a string 
     */
    @Override
    public String toString() { 
    	return value; 
    }
    
    /**
     * Return the authenticator associated to this type
     * 
     * @return an authenticator, shouldn't be null
     */
	public AuthenticatorInterface getAuthenticator(){
		return authenticators.get(this);
	}
	
	/**
	 * Return the name that can be used in the graphical interface 
	 * an eventually localized
	 * 
	 * @return a not null string
	 */
	public String getLocalizedName(){
		return value;
	}
}
  