/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.enumerations;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.jaspersoft.webservice.data.util.WebServiceConnector;
import com.jaspersoft.webservice.data.util.WebServiceConnectorInterface;

import net.sf.jasperreports.engine.JasperReportsContext;

/**
 * Class used to provide the verb types. It work in a similar way to an enum
 * but it is a class to be easily serialized with jackson. 
 * 
 * @author Orlandin Marco
 *
 */
public class VerbType {
	
	/**
	 * The logger
	 */
	protected static final Log logger = LogFactory.getLog(VerbType.class);
	
	/**
	 * The map where all the types instantiated (also if it is a subclass) are stored 
	 */
    private static final Map<String, VerbType> KINDS = new HashMap<String, VerbType>();
    
    /**
     * Map with the connector for each verb type
     */
    protected static final HashMap<VerbType, WebServiceConnectorInterface> connectors = new HashMap<VerbType, WebServiceConnectorInterface>();

    /**
     * Value to represent the verb post
     */
    public static final VerbType POST = new VerbType("post", new WebServiceConnector());
    
    /**
     * value to represent the verb get
     */
    public static final VerbType GET = new VerbType("get", new WebServiceConnector());

    /**
     * The value of the current type
     */
    private final String value;

    /**
     * Create a new type
     * 
     * @param value The value of the current type, should be unique, but this is not check to avoid to 
     * overwrite other types
     * @param connector the class that handle the communication with the server for this verb
     */
    protected VerbType(final String value, final WebServiceConnectorInterface connector) {
        if (KINDS.containsKey(value)) {
    		throw new RuntimeException("Connector for "+value+" already defined");
    	}
    	this.value = value;
        KINDS.put(value, this);
        connectors.put(this, connector);
    }

    /**
     * Get a verb type from its textual value
     * 
     * @param kind the textual representation of the type
     * @return the LanguageType associated or null if the name doesn't correspond to
     * any verb 
     */
    public static VerbType valueOf(final String kind) {
	    return KINDS.get(kind);
    }
    
    /**
     * Get an verb type from its textual value. If it is not found 
     * try to search it between the context extensions
     * 
     * @param kind the textual representation of the type
     * @param context a jasperreports context to search the authenticator between the extensions
     * if at first it can't be found
     * @return the VerbType associated or null if the name doesn't correspond to
     * any verb 
     */
    public static VerbType valueOf(final String kind, JasperReportsContext context) {
    	VerbType result = valueOf(kind);
	    //Not found, try to search it inside the extensions
	    if (result == null){
	    	context.getExtensions(VerbType.class);
	    	result = valueOf(kind);
	    }
	    return result;
    }

    /**
     * Return all the defined verb types
     * 
     * @return a not null list of the textual representation of the verb types
     */
    public static Collection<VerbType> values(){
    	return KINDS.values();
    }
    
    /**
     * Return all the defined verb types. Before try to load all the verbs
     * provided by the extension
     * 
     * @return a not null list of the textual representation of the verb types
     *
     */
    public static Collection<VerbType> values(JasperReportsContext context) {
    	context.getExtensions(VerbType.class);
    	return values();
    }
    
    /**
     * Return the current type as a string 
     */
    @Override
    public String toString() { 
    	return value; 
    }
    
    /**
     * Return the connector associated to this type
     * 
     * @return a WebServiceConnectorInterface, shouldn't be null
     */
	public WebServiceConnectorInterface getConnector(){
		return connectors.get(this);
	}
	
	/**
	 * Return the name that can be used in the graphical interface 
	 * an eventually localized
	 * 
	 * @return a not null string
	 */
	public String getLocalizedName(){
		return value;
	}
}
  