/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.json;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Date;
import java.util.Map;

import com.fasterxml.jackson.databind.JsonNode;
import com.jaspersoft.webservice.data.query.IWSDataSource;

import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JRField;
import net.sf.jasperreports.engine.data.JsonDataSource;
import net.sf.jasperreports.engine.util.JsonUtil;

/**
 * Extend the Json datasource to provide the expressions for the fields in a 
 * different way from the standard one: using a map field name\expression instead than
 * the field description. Since the visibility of some method doesn't easily allow this extension
 * some stuff is done trough the reflections. 
 * 
 * @author Orlandin Marco
 *
 */
public class WSJsonDataSource extends JsonDataSource implements IWSDataSource {
	
	/**
	 * Map with the expression for each field
	 */
	private Map<String, String> fieldPathMapper;

	/**
	 * The field that is accessed trough the reflection
	 */
	private Field currentNodeField = null;
	
	/**
	 * The current base path expression
	 */
	private String selectExpression;

	/**
	 * The binary data that represent the input of the datasource, this can be easily cloned
	 * when the a copy of the datasource is requested
	 */
	private ByteArrayOutputStream jsonStream;
	
	/**
	 * Convert a stream containing a json into a json node, that is the the root of 
	 * the content of the file. Every stream opened inside this method is closed
	 * 
	 * @param stream the stream containing the json data
	 * @return the full json structure inside the stream, this is the root of the structure
	 * @throws JRException if something goes wrong (ie the stream doesn't contain a json file).
	 */
	protected static JsonNode convertStreamIntoTree(ByteArrayOutputStream stream) throws JRException{
		ByteArrayInputStream inStream = new ByteArrayInputStream(stream.toByteArray());
		JsonNode node;
		try{
			 node = JsonUtil.parseJson(inStream);
		} catch (Exception ex){
			throw new JRException(ex);
		} finally {
			try {
				if (inStream != null) inStream.close();
			} catch (IOException e) {
			}
		}
		return node;
	}
	
	/**
	 * Create the datasource 
	 * 
	 * @param stream the stream to the json input, must be not null
	 * @param selectExpression the rootpath that is used to extract the subset of nodes used to get the values of the field, must be not null
	 * @param fieldPathMapper the map that associate each field name with its expression, must be not null
	 */
	public WSJsonDataSource(ByteArrayOutputStream jsonStream, String selectExpression, Map<String,String> fieldPathMapper) throws JRException {
		super(convertStreamIntoTree(jsonStream), selectExpression);
		this.fieldPathMapper =  fieldPathMapper;;
		this.selectExpression = selectExpression;
		this.jsonStream = jsonStream;
	}

	/**
	 * Return the expression for a specific field by getting it from the map
	 * 
	 * @param jrField the field
	 * @return a string representing the expression of the field
	 */
	protected String getExpressionValue(JRField jrField) {
		return fieldPathMapper.get(jrField.getName());
	}
	
	/**
	 * (non-Javadoc)
	 * 
	 * @see net.sf.jasperreports.engine.JRDataSource#getFieldValue(net.sf.jasperreports.engine.JRField)
	 */
	public Object getFieldValue(JRField jrField) throws JRException 
	{
		JsonNode currentJsonNode = getCurrentNode();
		if(currentJsonNode == null) {
			return null;
		}
		String expression = getExpressionValue(jrField);
		Object value = null;
		
		Class<?> valueClass = jrField.getValueClass();
		JsonNode selectedObject = getJsonData(currentJsonNode, expression);
		
		if(Object.class != valueClass) 
		{
			boolean hasValue = selectedObject != null 
					&& !selectedObject.isMissingNode() && !selectedObject.isNull();
			if (hasValue) 
			{
				try {
					if (valueClass.equals(String.class)) {
                        if (selectedObject.isArray()) {
                            value = selectedObject.toString();
                        } else {
                            value = selectedObject.asText();
                        }
					} else if (valueClass.equals(Boolean.class)) {
						value = selectedObject.booleanValue();
					} else if (Number.class.isAssignableFrom(valueClass)) {
						//FIXME if the json node is a number, avoid converting to string and parsing back the value
						value = convertStringValue(selectedObject.asText(), valueClass);	
					}
					else if (Date.class.isAssignableFrom(valueClass)) {
							value = convertStringValue(selectedObject.asText(), valueClass);
							
					} else {
						throw new JRException("Field '" + jrField.getName() + "' is of class '" + valueClass.getName() + "' and cannot be converted");
					}
				} catch (Exception e) {
					throw new JRException("Unable to get value for field '" + jrField.getName() + "' of class '" + valueClass.getName() + "'", e);
				}
			}
		}
		else
		{
			//GN this should fix the input controls
			value = selectedObject.asText();
		}
		
		return value;
	}
	
	/**
	 * Method used to access a private field of the superclass
	 * 
	 * @return the value of the private field
	 */
	protected JsonNode getCurrentNode(){
		try {
			if (currentNodeField == null){
				currentNodeField = JsonDataSource.class.getDeclaredField("currentJsonNode");
				currentNodeField.setAccessible(true);
			}
			return (JsonNode) currentNodeField.get(this);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * @see #com.jaspersoft.webservice.data.query.clone()
	 */
	@Override
	public IWSDataSource clone(){
		try {
			WSJsonDataSource clone = new WSJsonDataSource(jsonStream, selectExpression, fieldPathMapper);
			clone.setNumberPattern(getNumberPattern());
			clone.setDatePattern(clone.getDatePattern());
			return clone;
		} catch (JRException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * @see #com.jaspersoft.webservice.data.query.clone(String basePath, Map<String, String> fieldExpressionMap)
	 */
	@Override
	public IWSDataSource clone(String basePath, Map<String, String> fieldExpressionMap) {
		try {
			WSJsonDataSource clone = new WSJsonDataSource(jsonStream, basePath, fieldExpressionMap);
			clone.setNumberPattern(getNumberPattern());
			clone.setDatePattern(clone.getDatePattern());
			return clone;
		} catch (JRException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * @see #com.jaspersoft.webservice.data.query.clone(String basePath)
	 */
	@Override
	public IWSDataSource clone(String basePath) {
		try {
			WSJsonDataSource clone = new WSJsonDataSource(jsonStream, basePath, fieldPathMapper);
			clone.setNumberPattern(getNumberPattern());
			clone.setDatePattern(clone.getDatePattern());
			return clone;
		} catch (JRException e) {
			e.printStackTrace();
		}
		return null;
	}
}
