/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.jsonql;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.jaspersoft.webservice.data.query.IWSDataSource;
import com.jaspersoft.webservice.data.query.IWSQueryExecuterWrapper;
import com.jaspersoft.webservice.data.query.WebServiceQueryExecuterFactory;

import net.sf.jasperreports.engine.JRDataset;
import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JRValueParameter;
import net.sf.jasperreports.engine.JasperReportsContext;
import net.sf.jasperreports.engine.data.JsonQLDataSource;
import net.sf.jasperreports.engine.query.JsonQLQueryExecuter;
import net.sf.jasperreports.engine.query.JsonQueryExecuterFactory;

/**
 * A wrapper for the jsonQL query executer, essentially this override the creation of the data adapter
 * creating a subclass of the original one that work on a stream passed as parameter as input
 * 
 * @author Gianluca Natali
 *
 */
public class JsonQLQueryExecuterWrapper extends JsonQLQueryExecuter implements IWSQueryExecuterWrapper {
	
	/**
	 * Logger for the error
	 */
	protected final Log logger = LogFactory.getLog(JsonQLQueryExecuterWrapper.class);

	/**
	 * The datasource json that will extract the data from the input stream
	 */
	private WSJsonQLDataSource datasource;
	
	/**
	 * A map with the couple name\expression for each field
	 */
	private Map<String, String> fieldPathMap;
	
	/**
	 * The stream for the json datasource
	 */
	private ByteArrayOutputStream jsonInputStream = null;
	
	/**
	 * The rootpath for the json datasource
	 */
	private String rootPath;
	
	public JsonQLQueryExecuterWrapper(JasperReportsContext jasperReportsContext,
			JRDataset dataset,
			Map<String, ? extends JRValueParameter> parametersMap) {
		super(jasperReportsContext, dataset, parametersMap);
	}
	
	/**
	 * Create the WSJsonDataSource instead of a standard json datasource
	 */
	@Override
	public IWSDataSource createDatasource() throws JRException
	{
		if (jsonInputStream != null) {
			datasource = new WSJsonQLDataSource(jsonInputStream, getRootPath(),getFieldPathMap());

			//Set the pattern parameters if available
			JRValueParameter parameter = getValueParameter(WebServiceQueryExecuterFactory.NUMBER_PATTERN_NAME, true);
			if (parameter != null){
				datasource.setNumberPattern(parameter.getValue().toString());
			}
			parameter = getValueParameter(WebServiceQueryExecuterFactory.DATE_PATTERN_NAME, true);
			if (parameter != null){
				datasource.setDatePattern(parameter.getValue().toString());
			}
		} else {
			if (logger.isWarnEnabled()) {
				logger.warn("No JSON source was provided.");
			}
		}
		
		inherithedFromJsonQE(datasource);
		
		return datasource;
	}

	/**
	 * Set in the json datasource the parameter if they were provided
	 * 
	 * @param datasource a json datasource
	 */
	private void inherithedFromJsonQE(JsonQLDataSource datasource) {
		if (datasource != null) {
			String dateFormatPattern = getStringParameterOrProperty(JsonQueryExecuterFactory.JSON_DATE_PATTERN);
			if(dateFormatPattern != null){
				datasource.setDatePattern(dateFormatPattern);
			}
			
			String numberFormatPattern = getStringParameterOrProperty(JsonQueryExecuterFactory.JSON_NUMBER_PATTERN);
			if(numberFormatPattern != null){
				datasource.setNumberPattern(numberFormatPattern);
			}
			
			Locale jsonLocale = (Locale) getParameterValue(JsonQueryExecuterFactory.JSON_LOCALE, true);
			if (jsonLocale != null) {
				datasource.setLocale(jsonLocale);
			} else {
				String jsonLocaleCode = getStringParameterOrProperty(JsonQueryExecuterFactory.JSON_LOCALE_CODE);
				if (jsonLocaleCode != null) {
					datasource.setLocale(jsonLocaleCode);
				}
			}
			
			TimeZone jsonTimezone = (TimeZone) getParameterValue(JsonQueryExecuterFactory.JSON_TIME_ZONE, true);
			if (jsonTimezone != null) {
				datasource.setTimeZone(jsonTimezone);
			} else {
				String jsonTimezoneId = getStringParameterOrProperty(JsonQueryExecuterFactory.JSON_TIMEZONE_ID);
				if (jsonTimezoneId != null) {
					datasource.setTimeZone(jsonTimezoneId);
				}
			}
		}
	}
	
	/**
	 * Close the stream 
	 */
	@Override
	public void close() {
		super.close();
		closeStream();
	}
	
	protected void closeStream(){
		if (jsonInputStream != null){
			try {
				jsonInputStream.close();
				jsonInputStream = null;
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	//GETTERS AND SETTERS
	
	public Map<String, String> getFieldPathMap() {
		return fieldPathMap;
	}

	public void setFieldPathMap(Map<String, String> fieldPathMap) {
		this.fieldPathMap = fieldPathMap;
	}

	public String getRootPath() {
		return rootPath;
	}

	public void setRootPath(String rootPath) {
		this.rootPath = rootPath;
	}
	
	public ByteArrayOutputStream getJsonInputStream() {
		return jsonInputStream;
	}

	public void setJsonInputStream(ByteArrayOutputStream jsonInputStream) {
		closeStream();
		this.jsonInputStream = jsonInputStream;
	}

}
