/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.jsonql;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Map;

import com.fasterxml.jackson.databind.JsonNode;
import com.jaspersoft.webservice.data.query.IWSDataSource;

import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JRField;
import net.sf.jasperreports.engine.data.JsonQLDataSource;
import net.sf.jasperreports.engine.json.JRJsonNode;
import net.sf.jasperreports.engine.util.JsonUtil;
import net.sf.jasperreports.engine.util.json.JsonQLExecuter;

/**
 * Extend the JsonQL datasource to provide the expressions for the fields in a 
 * different way from the standard one: using a map field name\expression instead than
 * the field description. Since the visibility of some method doesn't easily allow this extension
 * some stuff is done trough the reflections. 
 * 
 * @author Orlandin Marco
 *
 */
public class WSJsonQLDataSource extends JsonQLDataSource implements IWSDataSource {
	
	/**
	 * Map with the expression for each field
	 */
	private Map<String, String> fieldPathMapper;

	/**
	 * The field that is accessed trough the reflection
	 */
	private Field currentNodeField = null;

	/**
	 * Other field that is accessed trough the reflection
	 */
	private Field jsonQLExecuterField = null;
	
	/**
	 * The current base path expression
	 */
	private String selectExpression;

	/**
	 * The binary data that represent the input of the datasource, this can be easily cloned
	 * when the a copy of the datasource is requested
	 */
	private ByteArrayOutputStream jsonStream;
	
	/**
	 * The root node of the json
	 */
	private JRJsonNode root;
	
	/**
	 * Convert a stream containing a json into a json node, that is the the root of 
	 * the content of the file. Every stream opened inside this method is closed
	 * 
	 * @param stream the stream containing the json data
	 * @return the full json structure inside the stream, this is the root of the structure
	 * @throws JRException if something goes wrong (ie the stream doesn't contain a json file).
	 */
	protected static JsonNode convertStreamIntoTree(ByteArrayOutputStream stream) throws JRException{
		ByteArrayInputStream inStream = new ByteArrayInputStream(stream.toByteArray());
		JsonNode node;
		try{
			 node = JsonUtil.parseJson(inStream);
		} catch (Exception ex){
			throw new JRException(ex);
		} finally {
			try {
				if (inStream != null) inStream.close();
			} catch (IOException e) {
			}
		}
		return node;
	}
	
	/**
	 * Create the datasource 
	 * 
	 * @param stream the stream to the json input, must be not null
	 * @param selectExpression the rootpath that is used to extract the subset of nodes used to get the values of the field, must be not null
	 * @param fieldPathMapper the map that associate each field name with its expression, must be not null
	 */
	public WSJsonQLDataSource(ByteArrayOutputStream jsonStream, String selectExpression, Map<String,String> fieldPathMapper) throws JRException {
		this(new JRJsonNode(null, convertStreamIntoTree(jsonStream)), selectExpression, fieldPathMapper);
		this.jsonStream = jsonStream;
	}
	
	/**
	 * Create the datasource 
	 * 
	 * @param root the root of the json
	 * @param selectExpression the rootpath that is used to extract the subset of nodes used to get the values of the field, must be not null
	 * @param fieldPathMapper the map that associate each field name with its expression, must be not null
	 */
	protected WSJsonQLDataSource(JRJsonNode root, String selectExpression, Map<String,String> fieldPathMapper) throws JRException {
		super(root, selectExpression);
		this.fieldPathMapper =  fieldPathMapper;
		this.selectExpression = selectExpression;
		this.root = root;
	}

	/**
	 * Return the expression for a specific field by getting it from the map
	 * 
	 * @param jrField the field
	 * @return a string representing the expression of the field
	 */
	protected String getExpressionValue(JRField jrField) {
		return fieldPathMapper.get(jrField.getName());
	}
	
    @Override
    public Object getFieldValue(JRField jrField) throws JRException {
    	JRJsonNode currentJsonNode = getCurrentNode();
    	
    	if(currentJsonNode == null) {
            return null;
        }

    	String expression = getExpressionValue(jrField);
        if (expression == null || expression.length() == 0) {
            return null;
        }

        JRJsonNode selectedNode = getQueryExecuter().selectNode(currentJsonNode, root, expression);
        if (selectedNode != null) {
            return getConvertedValue(selectedNode, jrField);
        }

        return null;
    }
	
	/**
	 * Method used to access a private field of the superclass
	 * 
	 * @return the value of the private field
	 */
	protected JRJsonNode getCurrentNode(){
		try {
			if (currentNodeField == null){
				currentNodeField = JsonQLDataSource.class.getDeclaredField("currentJsonNode");
				currentNodeField.setAccessible(true);
			}
			return (JRJsonNode) currentNodeField.get(this);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Method used to access a private field of the superclass
	 * 
	 * @return the value of the private field
	 */
	protected JsonQLExecuter getQueryExecuter(){
		try {
			if (jsonQLExecuterField == null){
				jsonQLExecuterField = JsonQLDataSource.class.getDeclaredField("jsonQLExecuter");
				jsonQLExecuterField.setAccessible(true);
			}
			return (JsonQLExecuter) jsonQLExecuterField.get(this);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * @see #com.jaspersoft.webservice.data.query.clone()
	 */
	@Override
	public IWSDataSource clone(){
		try {
			WSJsonQLDataSource clone = new WSJsonQLDataSource(jsonStream, selectExpression, fieldPathMapper);
			clone.setNumberPattern(getNumberPattern());
			clone.setDatePattern(clone.getDatePattern());
			return clone;
		} catch (JRException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * @see #com.jaspersoft.webservice.data.query.clone(String basePath, Map<String, String> fieldExpressionMap)
	 */
	@Override
	public IWSDataSource clone(String basePath, Map<String, String> fieldExpressionMap) {
		try {
			WSJsonQLDataSource clone = new WSJsonQLDataSource(jsonStream, basePath, fieldExpressionMap);
			clone.setNumberPattern(getNumberPattern());
			clone.setDatePattern(clone.getDatePattern());
			return clone;
		} catch (JRException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * @see #com.jaspersoft.webservice.data.query.clone(String basePath)
	 */
	@Override
	public IWSDataSource clone(String basePath) {
		try {
			WSJsonQLDataSource clone = new WSJsonQLDataSource(jsonStream, basePath, fieldPathMapper);
			clone.setNumberPattern(getNumberPattern());
			clone.setDatePattern(clone.getDatePattern());
			return clone;
		} catch (JRException e) {
			e.printStackTrace();
		}
		return null;
	}
}
