/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.query;

import java.util.HashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.jaspersoft.webservice.data.WebServiceDataAdapterService;
import com.jaspersoft.webservice.data.enumerations.AuthType;
import com.jaspersoft.webservice.data.enumerations.LanguageType;
import com.jaspersoft.webservice.data.enumerations.VerbType;
import com.jaspersoft.webservice.data.json.JsonQueryExecuterWrapper;
import com.jaspersoft.webservice.data.jsonql.JsonQLQueryExecuterWrapper;
import com.jaspersoft.webservice.data.query.definition.WebServiceQuery;
import com.jaspersoft.webservice.data.query.exception.JRParseException;
import com.jaspersoft.webservice.data.util.WebServiceConnector;
import com.jaspersoft.webservice.data.util.WebServiceConnectorInterface;
import com.jaspersoft.webservice.data.util.WsResponse;
import com.jaspersoft.webservice.data.xml.XmlQueryExecuterWrapper;

import net.sf.jasperreports.engine.JRDataset;
import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JRParameter;
import net.sf.jasperreports.engine.JRValueParameter;
import net.sf.jasperreports.engine.JasperReportsContext;
import net.sf.jasperreports.engine.query.JRAbstractQueryExecuter;

/**
 * Query executer for the WebService data adapter. This essentially does the request to the server
 * and depending on the language of the response it uses another specific query executer to extract
 * the informations from the response. When the query is executed
 * the json with the response from the webserver is stored inside the REPORT_PARAMETERS_MAP
 * with key JRParameter.REPORT_CONNECTION. This allow to pass it for example to subreport 
 * and rewind it without send another request to the server
 * 
 * @author Orlandin Marco & Gianluca Natali
 *
 */
public class WebServiceQueryExecuter extends JRAbstractQueryExecuter {
	
	/**
	 * The logger
	 */
	protected static final Log logger = LogFactory.getLog(WebServiceQueryExecuter.class);
	
	/**
	 * The language name
	 */
	public static final String CANONICAL_LANGUAGE = "WebServiceQuery";

	/**
	 * The response for the server that will be the input of an xml or json query executer.
	 */
	private WsResponse wsResponse;
	
	/**
	 * The executed query
	 */
	private WebServiceQuery webQLQuery;
	
	/**
	 * The URI of the web server
	 */
	private String wsURI;
	
	/**
	 * The language of the expected response
	 */
	private LanguageType langType;
	
	/**
	 * The verb of the request to the webserver
	 */
	private VerbType verbType;
	
	/**
	 * The authentication used on the webserver
	 */
	private AuthType authType;
	
	/**
	 * The parameters for the authentication
	 */
	private HashMap<String, String> authParameters;
	
	/**
	 * The internal query executer, used to handle the server response, it is 
	 * stored to avoid to recreate it each time
	 */
	private IWSQueryExecuterWrapper internalQueryExecutor;
	
	/**
	 * The parameters of the adapter, they are passed also to the internal query executer
	 */
	private Map<String, ? extends JRValueParameter> parametersMap;

	@SuppressWarnings("unchecked")
	public WebServiceQueryExecuter(JasperReportsContext jasperReportsContext,JRDataset dataset,Map<String, ? extends JRValueParameter> parametersMap) {
		super(jasperReportsContext, dataset, parametersMap);
		this.parametersMap = parametersMap;
		this.wsURI = (String)getParameterValue(WebServiceDataAdapterService.URI_KEY, true);
		
		//Load the authentication
		String authTypeName = (String)getParameterValue(WebServiceDataAdapterService.AUTH_TYPE_KEY, true);
		this.authType = AuthType.valueOf(authTypeName, jasperReportsContext);
		
		//Load the verb
		String verbTypeName = (String)getParameterValue(WebServiceDataAdapterService.VERB_KEY, true);
		this.verbType = VerbType.valueOf(verbTypeName, jasperReportsContext);
		
		this.authParameters = (HashMap<String, String>)getParameterValue(WebServiceDataAdapterService.AUTH_PARAMETERS_KEY, true);
		this.langType = (LanguageType)getParameterValue(WebServiceDataAdapterService.LANGUAGE_KEY, true);
		parseQuery();
	}
	
	@Override
	protected String getCanonicalQueryLanguage() {
		return CANONICAL_LANGUAGE;
	}
	
	@Override
	public String getQueryString() {
		return super.getQueryString();
	}

	/**
	 * Return a not null webservice datasource
	 */
	@Override
	public IWSDataSource createDatasource() throws JRException {

		String queryString = getQueryString();
		//If the query is not valid return an empty datasource
		if(queryString == null || queryString.trim().isEmpty()){
			return new WSEmptyDatasource();
		}
		
		webQLQuery = WebServiceParser.parseQuery(queryString);
		
		String contentType = webQLQuery.getOptions().getContentType();
		String content = webQLQuery.getQueryData();
		String baserUrl = webQLQuery.getOptions().getBaseUrl();
		Map<String, String> headerMap = webQLQuery.getHeadersMap();
		Map<String, String> getParameters = webQLQuery.getGETParameters();
		
		WebServiceConnectorInterface connector = null;
		if (verbType != null && verbType.getConnector() != null){
			connector = verbType.getConnector();
		} else {
			//Fallback on the default connector
			connector = new WebServiceConnector();
		}

		try{
			wsResponse = connector.callWS(wsURI, baserUrl, verbType, authType,
					contentType ,content, authParameters, 
					headerMap, getParameters);
			
		} catch(JRException ex){
			throw ex;
		} catch(Exception ex){
			throw new JRException(ex);
		} 
		try{
			IWSDataSource wrappedSource = getInternalQE().createDatasource();
	
			//Store the internal executer inside the parameter map
			JRValueParameter parameter = getValueParameter(JRParameter.REPORT_PARAMETERS_MAP, false);
			@SuppressWarnings("unchecked")
			Map<String,Object> parametersMap = (Map<String,Object>)parameter.getValue(); 
			parametersMap.put(JRParameter.REPORT_CONNECTION, wrappedSource);
			
			return wrappedSource;
		} catch(JRException ex){
			throw new JRParseException(ex, langType, wsResponse);
		} catch(Exception ex){
			throw new JRException(ex);
		}

	}

	/**
	 * When this is close reflect the method on the internal executer
	 */
	@Override
	public void close() {	
		try {
			IWSQueryExecuterWrapper qe = getInternalQE();
			if(qe==null){
				return;
			}
			qe.close();
		} catch (JRException e) {
			logger.error(e);
		}
	}

	/**
	 * When the query is cancelled reflect the method on the internal executer
	 */
	@Override
	public boolean cancelQuery() throws JRException {
		IWSQueryExecuterWrapper qe = getInternalQE();
		if(qe==null){
			return false;
		}
		return qe.cancelQuery();
	}

	@Override
	protected String getParameterReplacement(String parameterName) {
		return String.valueOf(getParameterValue(parameterName));
	}

	/**
	 * Return a specific query executer to handle the server response, it must be of type 
	 * IWSQueryExecuterWrapper to assure that the data source is an IWSDataSource, supporting
	 * the clone methods
	 * 
	 * @return a IWSQueryExecuterWrapper query executer for the server response
	 * @throws JRException raised if the server response is in a language different from JSON or XML
	 */
	protected IWSQueryExecuterWrapper getInternalQE() throws JRException {
		if (this.webQLQuery == null) {
			return null;
		}
		if (this.internalQueryExecutor == null) {
			if (langType == LanguageType.JSON) {
				JsonQueryExecuterWrapper jQe = new JsonQueryExecuterWrapper(getJasperReportsContext(), dataset, parametersMap);
				jQe.setFieldPathMap(webQLQuery.getFieldsMap());
				jQe.setJsonInputStream(wsResponse.getOutputStream());
				jQe.setRootPath(webQLQuery.getOptions().getRootPath());
				this.internalQueryExecutor = jQe;
			}else if (langType == LanguageType.XML) {
				XmlQueryExecuterWrapper xmlQe = new XmlQueryExecuterWrapper(getJasperReportsContext(), dataset, parametersMap);
				xmlQe.setFieldPathMap(webQLQuery.getFieldsMap());
				xmlQe.setXmlInputStream(wsResponse.getOutputStream());
				xmlQe.setRootPath(webQLQuery.getOptions().getRootPath());
				this.internalQueryExecutor = xmlQe;
			} else if (langType == LanguageType.JSONQL){
				JsonQLQueryExecuterWrapper jQe = new JsonQLQueryExecuterWrapper(getJasperReportsContext(), dataset, parametersMap);
				jQe.setFieldPathMap(webQLQuery.getFieldsMap());
				jQe.setJsonInputStream(wsResponse.getOutputStream());
				jQe.setRootPath(webQLQuery.getOptions().getRootPath());
				this.internalQueryExecutor = jQe;
			} else {
				if (langType != null) {
					throw new JRException("Language " + langType.toString() + " is not supported");
				} else {
					throw new JRException("The language is null, if you are using a subdataset you need to connect the parameters to the main dataset");
				}
			}
		}

		return this.internalQueryExecutor;
	}
}
