/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.util;

import java.io.ByteArrayInputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.HttpResponse;
import org.apache.http.client.fluent.Request;
import org.apache.http.client.fluent.Response;
import org.apache.http.entity.AbstractHttpEntity;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.message.BasicHeader;
import org.apache.http.protocol.HTTP;
import org.apache.http.util.EntityUtils;

import com.jaspersoft.webservice.data.authenitcator.AuthenticatorInterface;
import com.jaspersoft.webservice.data.enumerations.AuthType;
import com.jaspersoft.webservice.data.enumerations.VerbType;
import com.jaspersoft.webservice.data.query.exception.JRConnectionException;

import net.sf.jasperreports.engine.JRException;

/**
 * Utility class that do the connection be the webservice, send the request
 * and return the response. Used for standard get and post requests
 * 
 * @author Orlandin Marco
 *
 */
public class WebServiceConnector implements WebServiceConnectorInterface {

	/**
	 * The logger
	 */
	protected static final Log logger = LogFactory.getLog(WebServiceConnector.class);

	/**
	 * The uri of the web service
	 */
	protected String wsURI;
	
	/**
	 * The verb of the request
	 */
	protected VerbType verbType;
	
	/**
	 * The content type of the request
	 */
	protected String contentType;
	
	/**
	 * The request content
	 */
	protected String content;
	
	/**
	 * The authorization type for the request
	 */
	protected AuthType authType;
	
	/**
	 * the base url for the resource
	 */
	protected String baseURL;
	
	/**
	 * The map of the parameters used for the authentication
	 */
	protected Map<String, String> authMap;
	
	/**
	 * The map of headers to insert into the request
	 */
	protected Map<String, String> headerMap;
	
	/**
	 * The map of the parameters used for a get request
	 */
	protected Map<String, String> GETParameters;
	
	/**
	 * The encoding used by the get parameters
	 */
	protected static final String ENCODING = "UTF-8";
	
	/**
	 * Create the request, send it to the server and return the response
	 *
	 * @param wsURI The uri of the web service
	 * @param baseURL the base url for the resource
	 * @param verbType The verb of the request
	 * @param authType The authorization type for the request
	 * @param contentType  The content type of the request (used only with a post verb)
	 * @param content The content of the request (used only with a post verb)
	 * @param authMap The map of the parameters used for the authentication
	 * @param headerMap The map of headers to insert into the request
	 * @param GETParameters The map of the parameters used for a get request (used only with a GET verb)
	 * @return the response from the server, should be not null
	 * @throws JRException raised if something goes wrong with the server
	 */
	@Override
	public WsResponse callWS(String wsURI, String baseURL, VerbType verbType, 
			  AuthType authType, String contentType, String content,
			  Map<String, String> authMap, Map<String, String> headerMap, 
			  Map<String,String> GETParameters) throws JRException {
		
		//Initialize the fields
		this.verbType = verbType;
		this.content = content;
		this.contentType = contentType;
		this.authType = authType;
		this.wsURI = wsURI;
		this.baseURL = baseURL;
		this.authMap = authMap;
		this.headerMap = headerMap;
		this.GETParameters = GETParameters;
		Request req = createRequestForWs();
		Response resp = null;
		try {
			resp = req.execute();
		} catch (Exception e) {
			logger.error("Communication error "+ e.toString(), e);
			throw new JRConnectionException(e, req.toString());
		}
		if (resp != null) {
			int responseCode = -1;
			byte[] bytes = null;
			try {
				HttpResponse httpResponse = resp.returnResponse();
				responseCode = httpResponse.getStatusLine().getStatusCode();
				bytes = EntityUtils.toByteArray(httpResponse.getEntity());
				ByteArrayInputStream stream = new ByteArrayInputStream(bytes);
				String requestInfo = req.toString();
				
				WsResponse result = new WsResponse(responseCode, requestInfo, stream);
				return result;
			}  catch (Exception e) {
				throw new JRConnectionException(e, req.toString(), responseCode, bytes);
			}
		} else {
			throw new JRException("Null response with call " + req.toString());
		}
	}
	
	/**
	 * Create the request and return it
	 * 
	 * @return a not null request ready to be executed
	 */
	private Request createRequestForWs() throws JRException {
		try{
			Request req = null;
			if(verbType == VerbType.POST){
				req = createPostRequest();
			}else if(verbType == VerbType.GET){
				req = createGetRequest();
			}else{
				if (verbType != null){
					throw new JRException("Verb should be POST or GET. "+ verbType + " is not a valid verb");
				} else {
					throw new JRException("The verb can't be null");
				}
			}
			
			setHeaders(req);		
			setAuthenticationHeader(req);
				
			return req;
		}catch(Exception ex){
			throw new JRException("Error creating request: " + ex.toString(),ex);
		}
	}
	
	/**
	 * Append the base url to the server url, if the base url is not empty
	 * 
	 * @return the final url for the request
	 * @throws JRException 
	 * 
	 */
	protected String getRequestUrl() throws JRException {
		String url = wsURI.trim();
		
		if(baseURL != null && !baseURL.trim().isEmpty()){
			url += baseURL;
		}
		return url;
	}
	
	/**
	 * Create a post request for the server and return it 
	 * 
	 * @return a not null request ready to be executed
	 * 
	 */
	private Request createPostRequest() throws JRException  {
		
		Request req = Request.Post(getRequestUrl());
		//setting json object to post request.
	    AbstractHttpEntity entity;
		try {
			entity = new ByteArrayEntity(content.getBytes(ENCODING));
		} catch (UnsupportedEncodingException e) {
			throw new JRException("Unsupported encoding for request! ",e);
		}
	    entity.setContentType(new BasicHeader(HTTP.CONTENT_TYPE, contentType));
	    req.body(entity);
	    return req;
	}
	
	/**
	 * Create a get request for the server and return it 
	 * 
	 * @return a not null request ready to be executed
	 * @throws JRException 
	 *  
	 */
	private Request createGetRequest() throws JRException {
		String url = getRequestUrl();
		url = setAuthenticationURL(url);
		
		//Load parameters and encode them
		StringBuilder urlParameters = new StringBuilder();
		for(Entry<String,String> parameter : GETParameters.entrySet()){
			try{
				String encName = URLEncoder.encode(parameter.getKey(), ENCODING).replace("+", "%20");
				String encValue = URLEncoder.encode(parameter.getValue(), ENCODING).replace("+", "%20");
				urlParameters.append(encName);
				urlParameters.append("="); //$NON-NLS-1$
				urlParameters.append(encValue);
				urlParameters.append("&"); //$NON-NLS-1$
			} catch (Exception ex){
				logger.error(parameter, ex);
				ex.printStackTrace();
			}
		}
		
		String fromClause = urlParameters.toString();
		if (fromClause.endsWith("&")) {
			fromClause = fromClause.substring(0, fromClause.length()-1); 
		}
		if (!fromClause.isEmpty()){
			//if the URL doesn't have a question mark add it
			//if it has a question mark but it isn't the last char then there
			//are already other params encoded in the URL, so add so a "&" must
			//be the last symbol to separate it from the new parameters
			if(!url.contains("?")){
				url += "?";
			} else if (!url.endsWith("?") && !url.endsWith("&")){
				url += "&";
			}
			url += fromClause;
		}
		
		return Request.Get(url);
		
	}
	
	/**
	 * Set the authentication inside the request. It get authenticator
	 * for the authentication type and pass to it the authentication parameters
	 * and the request itself, and it will set the information inside the request
	 * 
	 * @param req the request where the authentication informations will be added
	 */
	protected void setAuthenticationHeader(Request req) {
		AuthenticatorInterface auth = authType.getAuthenticator();
		if (auth != null){
			auth.setAuthenication(req, authMap);
		}
	}
	
	/**
	 * Set the authentication on the URL. It get authenticator
	 * for the authentication type and pass to it the authentication parameters
	 * and the request itself, and it will set the information inside the request
	 * 
	 * @param URL the url of the request
	 * @param a not null and valid url that may contains new parameters
	 */
	private String setAuthenticationURL(String URL){
		AuthenticatorInterface auth = authType.getAuthenticator();
		if (auth != null){
			String newURL = auth.contributeURL(URL, authMap);
			if (newURL != null) return newURL; 
		}
		return URL;
	}

	/**
	 * Set the defined header inside the request
	 * 
	 * @param req request where the header are added
	 */
	protected void setHeaders(Request req) {
		for(Entry<String,String> header : headerMap.entrySet()){
			req.addHeader(header.getKey(), header.getValue());
		}
		
	}
}
