/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import org.apache.http.HttpStatus;

import net.sf.jasperreports.engine.JRException;

/**
 * Response returned by a server call
 * 
 * @author Orlandin Marco
 *
 */
public class WsResponse {

	/**
	 * The return code of the server. Even if the implementation of the connector
	 * could not use a standard http connection, the return code is assuming to
	 * use as standard the codes from the {@link HttpStatus}
	 */
	private int returnCode;
	
	/**
	 * The connection used to contact the server. This is used typically only for debug
	 * and logging, so it could be anything. It should describe what type of request was done
	 * on the server
	 */
	private String requestInfo;
	
	/**
	 * The response of the server as an array of byte
	 */
	private byte[] resultStream = new byte[0];

	public WsResponse(int returnCode, String requestInfo, InputStream resultStream) throws JRException{
		this.requestInfo = requestInfo;
		this.returnCode = returnCode;
		if (resultStream != null){
			ByteArrayOutputStream wsResponseStream = null;
			try{
				wsResponseStream = new ByteArrayOutputStream();
				byte[] buffer = new byte[1024];
				int len;
				while ((len = resultStream.read(buffer)) != -1 ) {
					wsResponseStream.write(buffer, 0, len);
				}
				wsResponseStream.flush();
				this.resultStream = wsResponseStream.toByteArray();
			} catch (Exception ex){
				throw new JRException(ex);
			} finally {
				if (wsResponseStream != null){
					try {
						wsResponseStream.close();
					} catch (IOException e) {
					}
				}
			}
		}
	}

	/**
	 * Get the return code 
	 * 
	 * @return a {@link HttpStatus} code. It is used only for debugging, however
	 * the ok code should be {@link HttpStatus.SC_OK}.
	 */
	public int getReturnCode() {
		return returnCode;
	}

	/**
	 * String that describe the connection with the server, used for debugging/logging
	 * 
	 * @return a String, could be null
	 */
	public String getRequestInfo() {
		return requestInfo;
	}

	/**
	 * Return a result of the call
	 * 
	 * @return return a not null array of byte
	 */
	public byte[] getResultStream() {
		return resultStream;
	}
	
	/**
	 * Create an output stream from the current response byte result. Each call generate
	 * a different stream
	 * 
	 * @return a not null {@link ByteArrayOutputStream}, must be closed by the user.
	 * 
	 */
	public ByteArrayOutputStream getOutputStream(){
		ByteArrayOutputStream baos = new ByteArrayOutputStream(resultStream.length);
		baos.write(resultStream, 0, resultStream.length);
		return baos;
	}
	
	/**
	 * Check if the response of the server has a valid code
	 * 
	 * @return true if the return code is fine, false otherwise
	 */
	public boolean isResponseValid(){
		return returnCode == HttpStatus.SC_OK;
	}
}
