/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.xml;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

import com.jaspersoft.webservice.data.query.IWSDataSource;

import net.sf.jasperreports.engine.DefaultJasperReportsContext;
import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JRField;
import net.sf.jasperreports.engine.data.JRXmlDataSource;
import net.sf.jasperreports.engine.util.JRXmlUtils;
import net.sf.jasperreports.engine.util.xml.JRXPathExecuter;
import net.sf.jasperreports.engine.util.xml.JRXPathExecuterUtils;

/**
 * Extend the xml datasource to provide the expressions for the fields in a 
 * different way from the standard one: using a map field name\expression instead than
 * the field description. Since the visibility of some method doesn't easily allow this extension
 * some stuff is done trough the reflections. 
 * 
 * @author Orlandin Marco
 *
 */
public class WSXmlDataSource extends JRXmlDataSource implements IWSDataSource {
	
	/**
	 * Map with the expression for each field
	 */
	private Map<String, String> fieldPathMapper;
	
	/**
	 * the xpath executer used internally
	 */
	protected JRXPathExecuter xPathExecuter;
	
	/**
	 * The binary data that represent the input of the datasource, this can be easily cloned
	 * when the a copy of the datasource is requested
	 */
	private ByteArrayOutputStream xmlStream;
	
	/**
	 * The current base path expression
	 */
	private String selectExpression;
	
	/**
	 * Convert a stream containing a xml into an xml document, that is the the root of 
	 * the content of the file. Every stream opened inside this method is closed
	 * 
	 * @param stream the stream containing the xml data
	 * @return the full xml structure inside the stream
	 * @throws JRException if something goes wrong (ie the stream doesn't contain a xml file).
	 */
	protected static Document convertStreamIntoDocument(ByteArrayOutputStream stream) throws JRException{
		ByteArrayInputStream inStream = new ByteArrayInputStream(stream.toByteArray());
		Document doc;
		try{
			doc = JRXmlUtils.parse(new InputSource(inStream), false);
		} catch (Exception ex){
			throw new JRException(ex);
		} finally {
			try {
				if (inStream != null) inStream.close();
			} catch (IOException e) {
			}
		}
		return doc;
	}
	
	/**
	 * Create the datasource 
	 * 
	 * @param stream the stream to the xml input, must be not null
	 * @param selectExpression the rootpath that is used to extract the subset of nodes used to get the values of the field, must be not null
	 * @param fieldPathMapper the map that associate each field name with its expression, must be not null
	 */
	public WSXmlDataSource(ByteArrayOutputStream xmlStream, String selectExpression, Map<String,String> fieldPathMapper) throws JRException {
		super(convertStreamIntoDocument(xmlStream), selectExpression);
		this.fieldPathMapper =  fieldPathMapper;
		this.xmlStream = xmlStream;
		this.selectExpression = selectExpression;
		//Create the xpath executer
		initXPathExecuter();
	}
	
	/**
	 * Try to get the xpath executer from the superclass trough the reflections, if 
	 * something goes wrong if create it like it is done in the superclass
	 */
	protected void initXPathExecuter() throws JRException{
		try {
			Field executerField = JRXmlDataSource.class.getDeclaredField("xPathExecuter");
			executerField.setAccessible(true);
			xPathExecuter = (JRXPathExecuter) executerField.get(this);
		} catch (Exception e) {
			xPathExecuter = JRXPathExecuterUtils.getXPathExecuter(DefaultJasperReportsContext.getInstance());
			e.printStackTrace();
		} 
	}

	/**
	 * Return the expression for a specific field by getting it from the map
	 * 
	 * @param jrField the field
	 * @return a string representing the expression of the field
	 */
	protected String getExpressionValue(JRField jrField) {
		return fieldPathMapper.get(jrField.getName());
	}
	
	/**
	 * (non-Javadoc)
	 * 
	 * @see net.sf.jasperreports.engine.JRDataSource#getFieldValue(net.sf.jasperreports.engine.JRField)
	 */
	public Object getFieldValue(JRField jrField) throws JRException 
	{
		Node currentNode = getCurrentNode();
		if(currentNode == null)
		{
			return null;
		}
		String expression = getExpressionValue(jrField);
		Object value = null;
		
		Class<?> valueClass = jrField.getValueClass();
		Object selectedObject = xPathExecuter.selectObject(currentNode, expression);

		if(Object.class != valueClass) 
		{
			if (selectedObject != null) 
			{
				if (selectedObject instanceof Node) 
				{
					String text = getText((Node) selectedObject);
					if (text != null) 
					{
						value = convertStringValue(text, valueClass);
					}
				} 
				else if (selectedObject instanceof Boolean && valueClass.equals(Boolean.class)) 
				{
					value = selectedObject;
				}
				else if (selectedObject instanceof Number && Number.class.isAssignableFrom(valueClass)) 
				{
					value = convertNumber((Number) selectedObject, valueClass);
				} 
				else 
				{
					String text = selectedObject.toString();
					value = convertStringValue(text, valueClass);
				}
			}
		}
		else
		{
			value = selectedObject;
		}
		return value;
	}

	/**
	 * @see #com.jaspersoft.webservice.data.query.clone()
	 */
	@Override
	public IWSDataSource clone(){
		try {
			WSXmlDataSource clone = new WSXmlDataSource(xmlStream, selectExpression, fieldPathMapper);
			clone.setDatePattern(getDatePattern());
			clone.setNumberPattern(getNumberPattern());
			return clone;
		} catch (JRException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * @see #com.jaspersoft.webservice.data.query.clone(String basePath, Map<String, String> fieldExpressionMap)
	 */
	@Override
	public IWSDataSource clone(String basePath, Map<String, String> fieldExpressionMap) {
		try {
			WSXmlDataSource clone = new WSXmlDataSource(xmlStream, basePath, fieldExpressionMap);
			clone.setDatePattern(getDatePattern());
			clone.setNumberPattern(getNumberPattern());
			return clone;
		} catch (JRException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * @see #com.jaspersoft.webservice.data.query.clone(String basePath)
	 */
	@Override
	public IWSDataSource clone(String basePath) {
		try {
			WSXmlDataSource clone = new WSXmlDataSource(xmlStream, basePath, fieldPathMapper);
			clone.setDatePattern(getDatePattern());
			clone.setNumberPattern(getNumberPattern());
			return clone;
		} catch (JRException e) {
			e.printStackTrace();
		}
		return null;
	}
}
