/*******************************************************************************
 * Copyright (C) 2005 - 2014 TIBCO Software Inc. All rights reserved.
 * http://www.jaspersoft.com.
 * 
 * Unless you have purchased  a commercial license agreement from Jaspersoft,
 * the following license terms  apply:
 * 
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/
package com.jaspersoft.webservice.data.xml;

import java.io.ByteArrayOutputStream;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.jaspersoft.webservice.data.query.IWSDataSource;
import com.jaspersoft.webservice.data.query.IWSQueryExecuterWrapper;
import com.jaspersoft.webservice.data.query.WebServiceQueryExecuterFactory;

import net.sf.jasperreports.engine.JRDataset;
import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JRValueParameter;
import net.sf.jasperreports.engine.JasperReportsContext;
import net.sf.jasperreports.engine.data.JRXmlDataSource;
import net.sf.jasperreports.engine.query.JRXPathQueryExecuterFactory;
import net.sf.jasperreports.engine.query.JsonQueryExecuter;

/**
 * A wrapper for the json query executer, essentially this override the creation of the data adapter
 * creating a subclass of the original one that work on a stream passed as parameter as input
 * 
 * @author Orlandin Marco
 *
 */
public class XmlQueryExecuterWrapper extends JsonQueryExecuter implements IWSQueryExecuterWrapper {
	
	/**
	 * Logger for the error
	 */
	protected final Log logger = LogFactory.getLog(XmlQueryExecuterWrapper.class);

	/**
	 * The datasource json that will extract the data from the input stream
	 */
	private WSXmlDataSource datasource;
	
	/**
	 * A map with the couple name\expression for each field
	 */
	private Map<String, String> fieldPathMap;
	
	/**
	 * The stream for the xml datasource
	 */
	private ByteArrayOutputStream xmlInputStream;
	
	/**
	 * The rootpath for the json datasource
	 */
	private String rootPath;
	
	public XmlQueryExecuterWrapper(JasperReportsContext jasperReportsContext, JRDataset dataset, Map<String, ? extends JRValueParameter> parametersMap) {
		super(jasperReportsContext, dataset, parametersMap);
	}
	
	@Override
	public IWSDataSource createDatasource() throws JRException
	{
		if (xmlInputStream != null) {
			datasource = new WSXmlDataSource(xmlInputStream, getRootPath(),getFieldPathMap());
			
			//Set the pattern parameters if available
			JRValueParameter parameter = getValueParameter(WebServiceQueryExecuterFactory.NUMBER_PATTERN_NAME, true);
			if (parameter != null){
				datasource.setNumberPattern(parameter.getValue().toString());
			}
			parameter = getValueParameter(WebServiceQueryExecuterFactory.DATE_PATTERN_NAME, true);
			if (parameter != null){
				datasource.setDatePattern(parameter.getValue().toString());
			}
		} else {
			if (logger.isWarnEnabled()) {
				logger.warn("No XML source was provided.");
			}
		}
		
		inherithedFromXMLQE(datasource);
		
		return datasource;
	}

	/**
	 * Set in the xml datasource the parameter if they were provided
	 * 
	 * @param datasource a xml datasource
	 */
	private void inherithedFromXMLQE(JRXmlDataSource datasource) {
		if (datasource != null) {
			String dateFormatPattern = getStringParameterOrProperty(JRXPathQueryExecuterFactory.XML_DATE_PATTERN);
			if(dateFormatPattern != null){
				datasource.setDatePattern(dateFormatPattern);
			}
			
			String numberFormatPattern = getStringParameterOrProperty(JRXPathQueryExecuterFactory.XML_NUMBER_PATTERN);
			if(numberFormatPattern != null){
				datasource.setNumberPattern(numberFormatPattern);
			}
			
			Locale jsonLocale = (Locale) getParameterValue(JRXPathQueryExecuterFactory.XML_LOCALE, true);
			if (jsonLocale != null) {
				datasource.setLocale(jsonLocale);
			}
			
			TimeZone jsonTimezone = (TimeZone) getParameterValue(JRXPathQueryExecuterFactory.XML_TIME_ZONE, true);
			if (jsonTimezone != null) {
				datasource.setTimeZone(jsonTimezone);
			}
		}
	}
	
	/**
	 * Close the stream 
	 */
	@Override
	public void close()
	{
		if(datasource != null){
			datasource.close();
		}
	}
	
	//GETTERS AND SETTERS

	public Map<String, String> getFieldPathMap() {
		return fieldPathMap;
	}

	public void setFieldPathMap(Map<String, String> fieldPathMap) {
		this.fieldPathMap = fieldPathMap;
	}
	
	/**
	 * If the rootpath is null or empty it return the path that point the root
	 * 
	 * @return a not null and not empty rootpath
	 */
	public String getRootPath() {
		return rootPath != null && !rootPath.isEmpty() ? rootPath : "/";
	}

	public void setRootPath(String rootPath) {
		this.rootPath = rootPath;
	}
	
	public ByteArrayOutputStream getXmlInputStream() {
		return xmlInputStream;
	}

	public void setXmlInputStream(ByteArrayOutputStream jsonInputStream) {
		this.xmlInputStream = jsonInputStream;
	}
}
